# <path to game directory>/addons/eventscripts/admin/libraries/modules.py
# by Adam Cunnington

"""Provide a simple way to create custom modules and register them to the 
core. Automatically handle action commands and menu interfaces.

"""

import os

from admin.information import CONSOLE_PREFIX, SHORTCUT_PREFIX, VERBOSE_NAME
from admin.libraries import commands, menus, tools

__all__ = (
    "ModulesError", 
    "module_manager", 
    "Module", 
    "Action", 
    "Parameter", 
    "Menu", 
    "DynamicMenu", 
)

_DEFAULT_INFORMER = tools.tell

class ModulesError(Exception):
    """General error encountered relating to admin.libraries.modules"""
    
class _ModuleManager(object):
    def __init__(self):
        self.menu = menus.Menu()
        self.menu.title = VERBOSE_NAME
        self.menu.description = ":: Module Administration"
        self.modules = {}
        self.actions = {}
        self.permissions = set()
        
module_manager = _ModuleManager()

class Module(object):
    """Create a module that represents a subset of functionality of the admin
    system core and register appropriate actions to reside in this module.
    
    """
    
    def __init__(self, basename, verbose_name=None):
        """Instantiate a Module object.
        
        Arguments:
        basename - the basename to identify the module by.
        verbose_name (Keyword Default: None) - the custom verbose name to use 
        in text.
        
        """
        if basename not in module_manager.modules:
            self.option = None
        else:
            if module_manager.modules[basename].loaded:
                raise ModulesError("%s is already loaded." % basename)
        self.basename = basename
        if verbose_name is None:
            self.verbose_name = tools.format_verbose_name(basename)
        else:
            self.verbose_name = verbose_name
        self.actions = set()
        self.menu = None
        self.loaded = True
        module_manager.modules[basename] = self
        
    def append(self, action):
        """Register an action to this module.
        
        Arguments:
        action - the action to register.
        
        """
        if self.menu is None:
            self.menu = menus.Menu(_module_menu_select)
            self.menu.title = "Module: %s" % self.verbose_name
            self.menu.description = ":: Actions"
            if self.option is None:
                self.option = menus.MenuOption(self.verbose_name)
                self.option.set_submenu(self.menu)
                module_manager.menu.append(action)
            else:
                self.option.text = self.verbose_name
                self.option.selectable = True
            if action.menu_interface:
                self.menu.append(action.option)
            self.actions.add(action)
            
    def unload(self):
        """Stop the module's administration functionality by unregistering all 
        registered actions and stopping any related menu items from being 
        selectable.
        
        """
        if self.menu is not None:
            self.option.selectable = False
            self.option.text = "UNLOADED: %s" % self.option.text
        for action in self.actions:
            action.unregister()
        self.loaded = False
        
    @property
    def path(self):
        return os.getcwd()
        
class Action(object):
    """Create an action to be listed in the administration system. 
    Functionality can vary from command handling to a fully functional chained 
    menu interface.
    
    """
    
    def __init__(self, basename, callback, menu_interface=False, 
                 server_command=False, requires_permission=True, 
                 verbose_name=None):
        """Instantiate an Action object.
        
        Arguments:
        basename - the basename to identify the action by.
        callback - the name to call when an action is submitted with valid 
        syntax and argument values.
        menu_interface (Keyword Default: False) - whether or not the action 
        should have a menu option and utilise the action's parameter's menus.  
        server_command (Keyword Default: False) - whether or not a server 
        command should be registered as well as client commands.
        requires_permission (Keyword Default: True) - whether or not there 
        should be a specific permission required to use this action.
        verbose_name (Keyword Default: None) - the custom verbose_name to use 
        in text.
        
        """
        self.basename = basename
        self.callback = callback
        self.menu_interface = menu_interface
        if requires_permission:
            permission = basename
        else:
            permission = None
        self._defaults = {}
        self._callback = commands.Callback(self._command_callback)
        client_console_command = commands.ClientConsoleCommand(self._callback)
        client_console_command.set_permission(permission)
        client_console_command.register("%s%s" %(CONSOLE_PREFIX, basename))
        client_say_command = commands.ClientSayCommand(self._callback)
        client_say_command.set_permission(permission)
        client_say_command.register("%s%s" %(SHORTCUT_PREFIX, basename))
        self.client_commands = (client_console_command, client_say_command)
        if not server_command:
            self.server_command = None
        else:
            self.server_command = commands.ServerCommand(self._callback)
            self.server_command.register("%s%s" %(CONSOLE_PREFIX, basename))
        if verbose_name is None:
            self.verbose_name = tools.format_verbose_name(basename)
        else:
            self.verbose_name = verbose_name
        if menu_interface:
            self.option = menus.MenuOption(self.verbose_name, self)
            self.option.set_permission(permission)
        if permission is not None:
            module_manager.permissions.add(permission)
        module_manager.actions[basename] = self
        
    def _command_callback(self, informer, *args, **kwargs):
        total_args = len(args)
        parameters = self._callback.parameters
        user_id = kwargs.pop("user_id")
        if (total_args == len(parameters) or 
            self.menu_interface is None):
            # If a full set of arguments have been passed or the action 
            # doesn't have a menu interface, then call the end callback.
            self._callback.callable_name(informer, user_id=user_id, *args)
            return
        # Otherwise, the next parameter's menu needs sending.
        parameters[total_args].menu.send(user_id, reset_navigation=False)
        
    def set_parameter(self, *args, **kwargs):
        """Register a parameter to be used by the action's command and 
        menu_interface if set.
        
        Arguments:
        args - the positional arguments defined by the commands.Callback class.
        kwargs - the keyword arguments defined by the commands.Callback class.
        
        """
        self._callback.set_parameter(*args, **kwargs)
        if self.menu_interface is None:
            # If there is no menu interface, then minimal functionaly should 
            # be assumed and so the parameter should function as that of a 
            # normal command parameter.
            return
        # Otherwise, all parameters should be declared as default to the 
        # commands library so that the callback is called regardless of 
        # whether compulsory parameters are omitted (as in this case, the menu 
        # needs to be used rather than raising an error).
        if self._callback.defaults:
            key, value = self._callback.defaults.popitem()
            self._defaults[key] = value
            
    def unregister(self):
        """Unregister all registered client and server commands."""
        for command in self.client_commands:
            command.unregister()
        if self.server_command is not None:
            self.server_command.unregister()
        del module_manager.actions[basename]
        
class Parameter(commands.Parameter):
    """Create a parameter that can be used across multiple actions."""
    
    def __init__(self, basename, description=None, item_getter=None, 
                 verbose_name=None):
        """Instantiate a Parameter object.
        
        Arguments:
        basename - the basename to identify the parameter by.
        description - a description of the values accepted.
        verbose_name (Keyword Default: None) - the custom verbose name to use 
        in text.
        item_getter - if the parameter's menu should be dynamic, the name to 
        call to dynamically fetch the menu's options.
        
        """
        super(Parameter, self).__init__(basename, description)
        if item_getter is None:
            self.menu = Menu()
        else:
            self.menu = DynamicMenu(item_getter)
        if verbose_name is None:
            self.verbose_name = tools.format_verbose_name(self.basename)
        else:
            self.verbose_name = verbose_name
        self.menu.title = self.verbose_name
        
class Menu(menus.Menu):
    """Create a menu to be used by a parameter, primarily or as a submenu."""
    
    def __init__(self):
        super(Menu, self).__init__()
        self.callback = _parameter_menu_select
        
class DynamicMenu(menus.DynamicMenu):
    """Create a dynamic menu to be used by a parameter, primarily or as a 
    submenu.
    
    """
    
    def __init__(self, item_getter):
        """Instantiate a DynamicMenu object.
        
        Arguments:
        item_getter - the name to call to dynamically fetch the menu's options.
        
        """
        super(DynamicMenu, self).__init__(item_getter)
        self.callback = _parameter_menu_select
        
def _module_menu_select(user_id, identifiers):
    action = identifiers[-1]
    parameters = action._callback.parameters
    if not parameters:
        # If the action has no parameters, then its callback can be called 
        # straight away.
        action._callback.callable_name(_DEFAULT_INFORMER, user_id=user_id)
        return
    # Otherwise, send the first parameter's menu.
    parameters[0].menu.send(user_id, reset_navigation=False)
    
def _parameter_menu_select(user_id, identifiers):
    action = module_manager.actions[identifiers.pop(0)]
    total_args = len(identifiers)
    parameters = action._callback.parameters
    if total_args < len(parameters):
        # If arguments have been omitted, the next parameter's menu needs 
        # sending.
        parameters[total_args].menu.send(user_id, reset_navigation=False)
    else:
        # Otherwise, the action's callback can be called.
        action._callback.callable_name(_DEFAULT_INFORMER, 
                                       user_id=user_id, *identifiers)